/*
 * Copyright (c) 1997 Stephen Williams (steve@icarus.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */


/*
 * This program prints information about the PCI devices it finds in
 * the /proc/pcibus directory. For this to work, you need to have
 * compiled into the kernel support for proc/pcibus.
 */

# include  <sys/types.h>
# include  <dirent.h>
# include  <stdio.h>
# include  <fcntl.h>
# include  <assert.h>

# include  "pcicode.h"

static void show_bar(__u32 bar)
{
      if (bar & 1) {
	    printf("    I/O registers at 0x%x", bar&~3);

      } else {
	    printf("    %sMemory", bar&0x08 ? "Prefetchable " : "");
	    switch (bar&6) {
		case 2:
		  printf(" (below 1meg)");
		  break;
		case 4:
		  printf(" (64bit address)");
		  break;
	    }

	    printf(" at 0x%08x\n", bar&~0xf);
      }
}

main()
{
      DIR*dir;
      struct dirent*de;

      dir = opendir("/proc/pcibus");
      if (! dir) {
	    fprintf(stderr, "Failed to open the /proc/pcibus directory.\n");
	    return 1;
      }

      for (de = readdir(dir) ;  de ;  de = readdir(dir)) {

	    unsigned bus_dev_fn;
	    __u32 space[64];
	    char path[128];
	    int fd, idx;
	    char*comma;

	    unsigned vend, devid;

	    if (de->d_name[0] == '.') continue;

	    bus_dev_fn = strtoul(de->d_name, 0, 16);

	    printf("Bus %u, Device %2u/%u", bus_dev_fn>>8,
		   (bus_dev_fn>>3) & 0x1f, bus_dev_fn & 0x7);

	    sprintf(path, "/proc/pcibus/%s", de->d_name);
	    fd = open(path, O_RDONLY, 0);
	    assert( fd >= 0 );
	    read(fd, space, sizeof space);
	    close(fd);

	    vend = space[0]&0xffff;
	    devid = (space[0]>>16) & 0xffff;

	    for (idx = 0; PciVenTable[idx].VenId < vend; idx += 1)
		    /* busy */;

	    if (PciVenTable[idx].VenId == vend) {
		  printf("    %s (0x%04x)\n", PciVenTable[idx].VenFull, vend);
	    } else {
		  printf("    Unknown vendor (0x%04x)\n", vend);
	    }


	    for (idx = 0; PciDevTable[idx].VenId < vend; idx += 1)
		    /* busy */;

	    while ((PciDevTable[idx].VenId == vend)
		   && (PciDevTable[idx].DevId != devid))
		  idx += 1;

	    if (PciDevTable[idx].VenId == vend) {
		  printf("    %s rev. %u (%s - 0x%04x)\n",
			 PciDevTable[idx].ChipDesc,
			 space[2]&0xff,
			 PciDevTable[idx].Chip, devid);
	    } else {
		  printf("    Unknown device rev %u (0x%04x)\n",
			 space[2]&0xff, devid);
	    }

	    printf("    %sMHZ Clock", space[1]&0x00200000? "66" : "33");
	    switch (space[1]&0x06000000) {
		case 0x00000000:
		  printf(", FAST Devsel");
		  break;
		case 0x02000000:
		  printf(", MEDIUM Devsel");
		  break;
		case 0x04000000:
		  printf(", SLOW Devsel");
		  break;
	    }

	    printf(", <");
	    comma="";
	    if (space[1]&0x00400000)
		  { printf("UDF"); comma=", "; }
	    if (space[1]&0x00800000)
		  { printf("%sfast back-to-back", comma); comma=", "; }
	    printf(">\n");


	    if ((space[2]&0xffffff00) == 0x06040000) {
		    /* The PCI-to-PCI bridge device is somewhat more
		       complicated then the standard type-0 device. It
		       has fewer BARs, but it has the various bridge
		       registers. */
		  for (idx = 0 ;  idx < 2 ;  idx += 1) if (space[idx+4])
			show_bar(space[idx+4]);

		  printf("    Bridge...\n");
		  printf("        memory from 0x%08x - 0x%08x\n",
			 (space[8]&0xfff0)<<16, space[8]|0xfffff);
		  printf("        I/O registers from 0x%04x - 0x%04x\n",
			 (space[7]&0xf0)<<8,
			 (space[7]&0xf000)|0x0fff);
		  if (space[9])
			printf("        prefetchable memory from "
			       "0x%08x - 0x%08x\n",
			       (space[9]&0xfff0)<<16, space[9]|0xfffff);
	    } else {
		  for (idx = 0 ;  idx < 6 ;  idx += 1) if (space[idx+4])
			show_bar(space[idx+4]);
	    }

      }

      return 0;
}
